/*

  yyycpu - a virtual machine
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "disassemble.hpp"

#include <vector>

mode_show_t mode_show=MODE_SHOW_UNSIGNEDDEC;
mode_address_t mode_address=MODE_ADDRESS_HEX;
bool mode_info_showunsigneddec=true;
bool mode_info_showsigneddec=true;
bool mode_info_showhex=true;
bool mode_info_showdist=true;
bool mode_codeaddress=true;
membunch *disas_membunch=0;
uint32 disas_addr=0;
bool disas_lastwasillegal=false;

template <class num_t>
static string unsigned2dec (const num_t &val)
{
  string r;
  const num_t zero(0);
  const num_t base(10);
  num_t c(val);
  while (c != zero) {
    r.insert (r.begin(), '0' + c % base);
    c /= base;
  };
  return (r.size()==0) ? "0" : r;
};

template <class num_t>
static string signed2dec (const num_t &val)
{
  if (val < num_t(0))
    return '-' + unsigned2dec(-val);
  else
    return unsigned2dec(val);
};

template <class num_t>
static string unsigned2hex (const num_t &val)
{
  string r;
  const num_t zero(0);
  const num_t base(16);
  num_t c(val);
  const char dec2hex[] = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};
  while (c != zero) {
    r.insert (r.begin(), dec2hex [c % base]);
    c /= base;
  };
  return (r.size()==0) ? "0" : r;
};

static string zeropad (const string &o, const uint8 count)
{
  string s(o);
  if (s.size()<count) s.insert(&s[0],count-s.size(),'0');
  return s;
};

static string show_uint8(const uint8 v)
{
  switch(mode_show) {
  case MODE_SHOW_UNSIGNEDDEC: return unsigned2dec(v) + 'n';
  case MODE_SHOW_SIGNEDDEC: return signed2dec(int8(v)) + 'n';
  case MODE_SHOW_HEX: return unsigned2hex(v) + 'h';
  };
  return string();
};

static string show_uint12(uint16 v)
{
  v&=0xfff;
  int16 i(v);
  if (i&0x800)i|=0xf000;
  switch(mode_show) {
  case MODE_SHOW_UNSIGNEDDEC: return unsigned2dec(v) + 'n';
  case MODE_SHOW_SIGNEDDEC: return signed2dec(i) + 'n';
  case MODE_SHOW_HEX: return unsigned2hex(v) + 'h';
  };
  return string();
};

static string show_uint16(const uint16 v)
{
  switch(mode_show) {
  case MODE_SHOW_UNSIGNEDDEC: return unsigned2dec(v) + 'n';
  case MODE_SHOW_SIGNEDDEC: return signed2dec(int16(v)) + 'n';
  case MODE_SHOW_HEX: return unsigned2hex(v) + 'h';
  };
  return string();
};

static string show_uint32(const uint32 v)
{
  switch(mode_show) {
  case MODE_SHOW_UNSIGNEDDEC: return unsigned2dec(v) + 'n';
  case MODE_SHOW_SIGNEDDEC: return signed2dec(int32(v)) + 'n';
  case MODE_SHOW_HEX: return unsigned2hex(v) + 'h';
  };
  return string();
};

static string info_uint8(const uint8 v)
{
  string r;
  if (mode_info_showunsigneddec) r += " = " + unsigned2dec(v)+'n';
  if (mode_info_showsigneddec) {
    int16 i=int8(v);
    if (i<0) r += " = " + signed2dec(i);
  };
  if (mode_info_showhex) r += " = "+ unsigned2hex(v)+'h';
  if (r.size()>=3) r.erase(&r[0],&r[3]);
  return r;
};

static string info_uint32(const uint32 v)
{
  string r;
  if (mode_info_showunsigneddec) r += " = " + unsigned2dec(v)+'n';
  if (mode_info_showsigneddec) {
    int32 i(v);
    if (i<0) r += " = " + signed2dec(i);
  };
  if (mode_info_showhex) r += " = "+ unsigned2hex(v)+'h';
  if (r.size()>=3) r.erase(&r[0],&r[3]);
  return r;
};

static string info_dist16(const int16 v, uint32 addr)
{
  if (!mode_info_showdist) return string();

  addr += int32(v) << 1;
  switch (mode_address) {
  case MODE_ADDRESS_HEX: return "PC := " + unsigned2hex(addr) + "h;";
  case MODE_ADDRESS_DEC: return "PC := " + unsigned2dec(addr) + "n;";
  };
  return string();
};

static string info_dist12(int16 v, uint32 addr)
{
  if (!mode_info_showdist) return string();

  v&=0xfff;
  if (v&0x800)v|=0xf000;
  addr += int32(v) << 1;
  switch (mode_address) {
  case MODE_ADDRESS_HEX: return "PC := " + unsigned2hex(addr) + "h;";
  case MODE_ADDRESS_DEC: return "PC := " + unsigned2dec(addr) + "n;";
  };
  return string();
};

static const string align(const vector<string> &s,const vector<int> &p)
{
  string r;
  int curpos=0;
  for (unsigned i=0;i<s.size();++i) {
    while (curpos<p[i]) {
      r+=' ';
      curpos++;
    };
    r+=s[i];
    curpos+=s[i].size();
  };
  while (r.size() && (r[r.size()-1]==' ')) r.erase(r.size()-1,r.size());
  return r;
};

static const string tg(const uint32 addr, const string &command, const string &p1,
		       const string &p2, const string &p3, const string &comment)
{
  vector<string> s;
  if (mode_codeaddress) {
    switch(mode_address) {
    case MODE_ADDRESS_HEX: s.push_back(zeropad(unsigned2hex(addr),8)+'h'); break;
    case MODE_ADDRESS_DEC: s.push_back(zeropad(unsigned2dec(addr),10)+'n'); break;
    };
  };
  s.push_back(command+' ');
  s.push_back(p1.empty() ? string() : p1 + (p2.empty()?string():string(",")));
  s.push_back(p2.empty() ? string() : p2 + (p3.empty()?string():string(",")));
  s.push_back(p3.empty() ? string() : p3+' ');
  s.push_back(comment.empty() ? string() : "; "+comment);
  vector<int> p;
  int a;
  if (mode_codeaddress) {
    switch(mode_address) {
    case MODE_ADDRESS_HEX: a=11; break;
    case MODE_ADDRESS_DEC: a=13; break;
    };
    p.push_back(0);
  }else{
    a=8;
  };
  p.push_back(a);
  p.push_back(a+8);
  p.push_back(a+16);
  p.push_back(a+24);
  p.push_back(a+40);
  return align(s,p);
};

static const string reg(const uint8 n)
{
  return 'R' + unsigned2dec(n);
};

const string disassemble ()
{
  disas_lastwasillegal=false;
  uint32 a=disas_addr;
  uint32 &d(disas_addr);
  membunch &m(*disas_membunch);

  uint16 ins=m.get16(a);
  switch (nib3(ins)) {
  case B0000:
    switch (nib2(ins)) {
    case B0000:
      switch(nib1(ins)) {
      case B0010: d+=6; return tg(a,"CONSD",reg(nib0(ins)),show_uint32(m.get32(a+2)),"",info_uint32(m.get32(a+2)));
      }; break;
    case B0010: d+=2; return tg(a,"OUT",reg(nib1(ins)),reg(nib0(ins)),"","");
    case B0011: d+=2; return tg(a,"INC",reg(nib1(ins)),reg(nib0(ins)),"","");
    case B0101: d+=2; return tg(a,"ADD",reg(nib1(ins)),reg(nib0(ins)),"","");
    case B0110: d+=2; return tg(a,"SUB",reg(nib1(ins)),reg(nib0(ins)),"","");
    }; break;
  case B0001:
    switch (nib2(ins)) {
    case B0001: d+=2; return tg(a,"MOV",reg(nib1(ins)),reg(nib0(ins)),"","");
    case B0110: d+=4; return tg(a,"XSGCW",reg(nib1(ins)),reg(nib0(ins)),show_uint16(m.get16(a+2)),info_dist16(m.get16(a+2),a));
    case B1100: d+=4; return tg(a,"XNECW",reg(nib1(ins)),reg(nib0(ins)),show_uint16(m.get16(a+2)),info_dist16(m.get16(a+2),a));
    }; break;
  case B0010:
    switch (nib2(ins) & B1110) {
    case B0100: d+=2; return tg(a,"SARC",reg(nib0(ins)),show_uint8((ins>>4)&31),"","");
    }; break;
  case B1011: d+=2; return tg(a,"JMPCT",show_uint12(ins & 0xfff),"","",info_dist12(ins&0xfff,a));
  case B1111: d+=2; return tg(a,"CONSB",reg(nib2(ins)),show_uint8(byt0(ins)),"",info_uint8(byt0(ins)));
  };
  disas_lastwasillegal=true; d+=2; return tg(a,"dw "+zeropad(unsigned2hex(ins),4)+'h',"","","","illegal instruction");
};
