/*

  yyycpu - a virtual machine
  Copyright (C) 2000  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "disassemble.hpp"
#include "../common/fatal.hpp"

#include <unistd.h>
#include <stdio.h>

void usage()
{
  printf(
	 "\n"
	 "Usage: yyydisas [options ...] input_file [output_file]\n"
	 "\n"
	 "if output_file is omitted, stdout is used\n"
	 "\n"
	 "options:\n"
	 "  -c       show the current address of each command\n"
	 "  -d       show addresses in decimal instead of hexadecimal\n"
	 "  -u       don't show unsigned decimal values in the comment\n"
	 "  -s       don't show signed decimal values in the comment\n"
	 "  -x       don't show hexadecimal values in the comment\n"
	 "  -j       don't show jumptarget address in the comment\n"
	 "  -e(ush)  encode parameters as unsigned, signed or hex\n"
	 "  -i       stop processing at the first illegal instruction\n"
	 "  -f val   start processing at address 'val'\n"
	 "  -t val   terminate processing before address 'val'\n"
	 "  -h       show this help screen\n"
	 "\n"
	 );
  exit(1);
};

static bool illegalstop=false;
static uint32 termaddr=0xFFFFFFFF;
FILE *input=0, *output=0;

int main(int argc, char *argv[])
{
  mode_codeaddress=false;
  int r;
  while ((r=getopt(argc,argv,"cdusxje:if:t:h"))!=-1) {
    switch (r) {
    case 'c': mode_codeaddress=true; break;
    case 'd': mode_address=MODE_ADDRESS_DEC; break;
    case 'u': mode_info_showunsigneddec=false; break;
    case 's': mode_info_showsigneddec=false; break;
    case 'x': mode_info_showhex=false; break;
    case 'j': mode_info_showdist=false; break;
    case 'e':
      if (strlen(optarg)!=1) fatal("only one character is allowed as parameter to -e option");
      switch (*optarg) {
      case 'u': mode_show=MODE_SHOW_UNSIGNEDDEC; break;
      case 's': mode_show=MODE_SHOW_SIGNEDDEC; break;
      case 'h': mode_show=MODE_SHOW_HEX; break;
      default: fatal("illegal parameter to -e option");
      };break;
    case 'i': illegalstop=true; break;
    case 'f': disas_addr=atoll(optarg); break;
    case 't': termaddr=atoll(optarg); break;
    case 'h': usage();
    case '?': usage();
    };
  };
  if (optind<argc) {
    input=fopen(argv[optind],"r");
    if (!input) fatal("cannot open input file");
    optind++;
  }else usage();
  if (optind<argc) {
    output=fopen(argv[optind],"w");
    if (!output) fatal("cannot open output file");
    optind++;
  }else{
    output=stdout;
  };
  if (optind!=argc) fatal("too many parameters on commandline");

  fseek(input,0,SEEK_END);
  if (ferror(input)) fatal("fseek end failed");
  uint32 len=ftell(input);
  fseek(input,0,SEEK_SET);
  if (ferror(input)) fatal("fseek start failed");
  uint8 *data = new uint8[len];
  fread(data,1,len,input);
  if (ferror(input)) fatal("fread failed");

  {
    membunch m(data,len);
    disas_membunch=&m;
    if (termaddr>len)termaddr=len;
    while (disas_addr<termaddr) {
      string s(disassemble());
      if (disas_lastwasillegal && illegalstop) break;
      fprintf(output,"%s\n",s.c_str());
    };
  };

  delete[] data;

  fclose(input);
  if (output!=stdout) fclose(output);

  return 0;
};
